@@ -3,6 +3,7 @@
 Remarks:
     This module is used primarily for the creation of simulations from data sets.
 
+TODO Decide where the Batchers should go
 TODO InferredEncoder needs a lot of love. Currently its inferences are quite bad.
 """
 
@@ -10,7 +11,7 @@ import json
 
 from collections import defaultdict
 from abc import ABC, abstractmethod
-from typing import Iterator, Sequence, Generic, TypeVar, Any, Optional, Hashable, Union, Dict, Tuple
+from typing import Iterator, Optional, Sequence, Generic, TypeVar, Any, Hashable, Union, Dict, Tuple
 
 T_out = TypeVar('T_out', bound=Hashable, covariant=True) 
 
@@ -38,6 +39,7 @@ class Encoder(Generic[T_out], ABC):
         if json_val == "onehot"  : return OneHotEncoder()
         if json_val == "string"  : return StringEncoder()
         if json_val == "inferred": return InferredEncoder()
+        if json_val == "factor"  : return FactorEncoder()
 
         raise Exception('We were unable to determine the appropriate encoder from json')
 
@@ -213,7 +215,7 @@ class NumericEncoder(Encoder[float]):
 class OneHotEncoder(Encoder[Tuple[int,...]]):
     """An Encoder implementation that turns incoming values into a one hot representation."""
 
-    def __init__(self, fit_values: Sequence[Any] = [], singular_if_binary: bool = True, error_if_unknown = False) -> None:
+    def __init__(self, fit_values: Sequence[Any] = [], singular_if_binary: bool = False, error_if_unknown = True) -> None:
         """Instantiate a OneHotEncoder.
 
         Args:
@@ -308,7 +310,7 @@ class FactorEncoder(Encoder[int]):
     """An Encoder implementation that turns incoming values into factor representation."""
 
     def __init__(self, fit_values: Sequence[Any] = [], error_if_unknown = False) -> None:
-        """Instantiate a OneHotEncoder.
+        """Instantiate a FactorEncoder.
 
         Args:
             fit_values: Provide the universe of values for encoding and set `is_fit==True`.
@@ -381,7 +383,7 @@ class FactorEncoder(Encoder[int]):
         try:
             return [ self._levels[value] for value in values ]
         except KeyError as e:
-            raise Exception(f"We were unable to find {e} in {self._levels.keys()}")
+            raise Exception(f"We were unable to find {e} in {self._levels.keys()}") from None
 
 class InferredEncoder(Encoder[Hashable]):
     """An Encoder implementation that looks at its given `fit` values and infers the best Encoder."""
@@ -422,36 +424,13 @@ class InferredEncoder(Encoder[Hashable]):
 
         raise Exception("This encoder must be fit before it can be used.")
 
-T_ignore  = TypeVar('T_ignore' , bound = Optional[bool]   , covariant=True) 
-T_label   = TypeVar('T_label'  , bound = Optional[bool]   , covariant=True) 
-T_encoder = TypeVar('T_encoder', bound = Optional[Encoder], covariant=True)
+T_ignore  = TypeVar('T_ignore' ) 
+T_label   = TypeVar('T_label'  ) 
+T_encoder = TypeVar('T_encoder')
 
 class Metadata(Generic[T_ignore, T_label, T_encoder]):
     """A storage class for Optional meta information describing features."""
 
-    @staticmethod
-    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'Metadata':
-        """Construct a Metadata object from JSON.
-
-        Args:
-            json_val: Either a json string or the decoded json object.
-
-        Returns:
-            The Metadata representation of the given JSON string or object.
-        """
-
-        config = json.loads(json_val) if isinstance(json_val,str) else json_val
-
-        ignore  = None if "ignore"   not in config else config["ignore"]
-        label   = None if "label"    not in config else config["label" ]
-        encoder = None if "encoding" not in config else Encoder.from_json(config["encoding"])
-
-        return Metadata(ignore,label,encoder)
-
-    @staticmethod
-    def default() -> 'Metadata[bool,bool,Encoder]':
-        return Metadata(False,False,InferredEncoder())
-
     @property
     def ignore(self) -> T_ignore:
         return self._ignore
@@ -465,7 +444,6 @@ class Metadata(Generic[T_ignore, T_label, T_encoder]):
         return self._encoder
 
     def __init__(self, ignore: T_ignore, label: T_label, encoder: T_encoder) -> None:
-        ...
         """Instantiate PartialMeta.
 
         Args:
@@ -500,4 +478,69 @@ class Metadata(Generic[T_ignore, T_label, T_encoder]):
         label   = override.label   if override.label   is not None else self.label
         encoder = override.encoder if override.encoder is not None else self.encoder
 
-        return Metadata[T_ignore, T_label, T_encoder](ignore, label, encoder)
\ No newline at end of file
+        return Metadata(ignore, label, encoder)
+
+class PartMeta(Metadata[Optional[bool], Optional[bool], Optional[Encoder]]):
+    """A storage class for partial meta information describing features."""
+
+    @staticmethod
+    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'PartMeta':
+        """Construct a PartMeta object from JSON.
+
+        Args:
+            json_val: Either a json string or the decoded json object.
+
+        Returns:
+            The PartMeta representation of the given JSON string or object.
+        """
+
+        config = json.loads(json_val) if isinstance(json_val,str) else json_val
+
+        ignore  = None if "ignore"   not in config else config["ignore"]
+        label   = None if "label"    not in config else config["label" ]
+        encoder = None if "encoding" not in config else Encoder.from_json(config["encoding"])
+
+        return PartMeta(ignore,label,encoder)
+
+    def __init__(self, ignore: bool = None, label: bool = None, encoder: Encoder = None):
+        """Instantiate FullMeta.
+
+        Args:
+            ignore: Indicates if the feature should be ignored.
+            label: Indicates if the feature should be regarded as a supervised label
+            encoder: The Encoder that should be used when ingesting features.
+        """
+
+        super().__init__(ignore, label, encoder)
+
+class FullMeta(Metadata[bool, bool, Encoder]):
+    """A storage class for full meta information describing features."""
+
+    @staticmethod
+    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'FullMeta':
+        """Construct a FullMeta object from JSON.
+
+        Args:
+            json_val: Either a json string or the decoded json object.
+
+        Returns:
+            The FullMeta representation of the given JSON string or object.
+        """
+
+        part_meta = PartMeta.from_json(json_val)
+
+        if part_meta.ignore is not None and part_meta.label is not None and part_meta.encoder is not None:
+            return FullMeta(part_meta.ignore, part_meta.label, part_meta.encoder)
+        
+        raise Exception("FullMeta JSON must define 'ignore', 'label' and 'encoder'")
+
+    def __init__(self, ignore: bool = False, label: bool = False, encoder: Encoder = StringEncoder()):
+        """Instantiate FullMeta.
+
+        Args:
+            ignore: Indicates if the feature should be ignored.
+            label: Indicates if the feature should be regarded as a supervised label
+            encoder: The Encoder that should be used when ingesting features.
+        """
+
+        super().__init__(ignore, label, encoder)
\ No newline at end of file
