@@ -7,27 +7,25 @@ Rather, they simply make it possible to use static type checking for any project
 to do so.
 
 TODO Add RegressionSimulation
+TODO Figure out LazySimulation vs JsonSimulation
 """
 
 import gc
 import csv
 import json
-import urllib.request
-import gzip
 
 from collections import defaultdict
 from itertools import compress, repeat, count, chain
-from http.client import HTTPResponse
-from contextlib import closing
 from abc import ABC, abstractmethod
-from hashlib import md5
 from typing import (
-    Optional, Iterable, Sequence, List, Union, Callable, TypeVar, 
-    Generic, Hashable, Dict, cast, Any, Tuple
+    Optional, Iterable, Sequence, List, Union, Callable, 
+    TypeVar, Generic, Hashable, Dict, Any, Tuple
 )
 
-from coba import random as cb_random
-from coba.preprocessing import FactorEncoder, Metadata, OneHotEncoder, NumericEncoder, Encoder
+import coba.random as cb_random
+
+from coba.data import HttpSource, DiskSource
+from coba.preprocessing import FactorEncoder, FullMeta, PartMeta, OneHotEncoder, NumericEncoder, Encoder
 from coba.execution import ExecutionContext
 
 Context = Optional[Hashable]
@@ -42,7 +40,9 @@ _A_out = TypeVar('_A_out', bound=Action, covariant=True)
 class Interaction(Generic[_C_out, _A_out]):
     """A class to contain all data needed to represent an interaction in a bandit simulation."""
 
-    def __init__(self, context: _C_out, actions: Sequence[_A_out], key: Key = 0) -> None:
+    #this is a problem with pylance compaining about covariance in constructor so we have to type ignore it. 
+    #See this ticket in mypy for more info https://github.com/python/mypy/issues/2850
+    def __init__(self, context: _C_out, actions: Sequence[_A_out], key: Key = 0) -> None: #type: ignore
         """Instantiate Interaction.
 
         Args
@@ -75,33 +75,6 @@ class Interaction(Generic[_C_out, _A_out]):
 class Simulation(Generic[_C_out, _A_out], ABC):
     """The simulation interface."""
 
-    @staticmethod
-    def from_json(json_val:Union[str, Dict[str, Any]]) -> 'Simulation[Context, Action]':
-        """Construct a Simulation object from JSON.
-
-        Args:
-            json_val: Either a json string or the decoded json object.
-
-        Returns:
-            The Simulation representation of the given JSON string or object.
-        """
-
-        config = json.loads(json_val) if isinstance(json_val,str) else json_val
-
-        no_shuffle  : Callable[[Simulation[Context, Action]], Simulation] = lambda sim: sim
-        seed_shuffle: Callable[[Simulation[Context, Action]], Simulation] = lambda sim: ShuffleSimulation(sim, config["seed"])
-
-        lazy_loader: Callable[[Callable[[],Simulation[Context,Action]]], Simulation] = lambda sim_factory: LazySimulation(sim_factory)
-        now_loader : Callable[[Callable[[],Simulation[Context,Action]]], Simulation]  = lambda sim_factory: sim_factory()
-
-        shuffler = seed_shuffle if "seed" in config and config["seed"] is not None else no_shuffle
-        loader   = lazy_loader  if "lazy" in config and config["lazy"] == True     else now_loader
-
-        if config["type"] == "classification":
-            return  loader(lambda: shuffler(ClassificationSimulation.from_json(config["from"])))
-
-        raise Exception("We were unable to recognize the provided simulation type")
-
     @property
     @abstractmethod
     def interactions(self) -> Sequence[Interaction[_C_out, _A_out]]:
@@ -113,7 +86,7 @@ class Simulation(Generic[_C_out, _A_out], ABC):
             it would be possible for it to only allow enumeration one time.
         """
         ...
-    
+
     @abstractmethod
     def rewards(self, choices: Sequence[Tuple[Key,Choice]] ) -> Sequence[Reward]:
         """The observed rewards for interactions (identified by its key) and their selected action indexes.
@@ -127,6 +100,62 @@ class Simulation(Generic[_C_out, _A_out], ABC):
         """
         ...
 
+class JsonSimulation(Simulation[Context, Action]):
+    """A Simulation implementation which supports loading and unloading from json representations.""" 
+    
+    def __init__(self, json_val) -> None:
+        """Instantiate a JsonSimulation
+
+        Args:
+            json: A json representation that can be turned into a simulation when needed.
+        """
+
+        self._json_obj = json.loads(json_val) if isinstance(json_val,str) else json_val
+        self._simulation: Optional[Simulation[Context, Action]]  = None
+
+    def __enter__(self) -> 'JsonSimulation':
+        """Load the simulation into memory. If already loaded do nothing."""
+
+        
+        if self._simulation is None and self._json_obj["type"] == "classification":
+            self._simulation = ClassificationSimulation.from_json(self._json_obj["from"])
+        else:
+            raise Exception("We were unable to recognize the provided simulation type")
+
+        return self
+
+    def __exit__(self, exception_type, exception_value, traceback) -> None:
+        """Unload the simulation from memory."""
+
+        if self._simulation is not None:
+            self._simulation = None
+            gc.collect() #in case the simulation is large
+
+    @property
+    def interactions(self) -> Sequence[Interaction[Context,Action]]:
+        """The interactions in this simulation.
+
+        Remarks:
+            See the Simulation base class for more information.
+        """
+
+        if self._simulation is not None:
+            return self._simulation.interactions
+        
+        raise Exception("A JsonSimulation must be loaded before it can be used.")
+
+    def rewards(self, choices: Sequence[Tuple[Key,Choice]]) -> Sequence[Reward]:
+        """The observed rewards for interactions (identified by its key) and their selected action indexes.
+
+        Remarks:
+            See the Simulation base class for more information.
+        """
+        
+        if self._simulation is not None:
+            return self._simulation.rewards(choices)
+
+        raise Exception("A JsonSimulation must be loaded before it can be used.")
+
 class LazySimulation(Simulation[_C_out, _A_out]):
     """A Simulation implementation which supports loading and unloading from memory.""" 
     
@@ -148,15 +177,13 @@ class LazySimulation(Simulation[_C_out, _A_out]):
 
             return self
 
-    def __exit__(self, exception_type, exception_value, traceback) -> 'LazySimulation[_C_out,_A_out]':
+    def __exit__(self, exception_type, exception_value, traceback) -> None:
         """Unload the simulation from memory."""
 
         if self._simulation is not None:
             self._simulation = None
             gc.collect() #in case the simulation is large
 
-        return self
-
     @property
     def interactions(self) -> Sequence[Interaction[_C_out,_A_out]]:
         """The interactions in this simulation.
@@ -167,7 +194,7 @@ class LazySimulation(Simulation[_C_out, _A_out]):
 
         if self._simulation is not None:
             return self._simulation.interactions
-        
+
         raise Exception("A LazySimulation must be loaded before it can be used.")
 
     def rewards(self, choices: Sequence[Tuple[Key,Choice]]) -> Sequence[Reward]:
@@ -186,7 +213,7 @@ class MemorySimulation(Simulation[_C_out, _A_out]):
     """A Simulation implementation created from in memory sequences of contexts, actions and rewards."""
 
     def __init__(self, 
-        contexts: Sequence[_C_out], 
+        contexts   : Sequence[_C_out], 
         action_sets: Sequence[Sequence[_A_out]], 
         reward_sets: Sequence[Sequence[Reward]]) -> None:
         """Instantiate a MemorySimulation.
@@ -233,9 +260,9 @@ class LambdaSimulation(Simulation[_C_out, _A_out]):
 
     def __init__(self,
                  n_interactions: int,
-                 context: Callable[[int],_C_out],
+                 context   : Callable[[int],_C_out],
                  action_set: Callable[[_C_out],Sequence[_A_out]], 
-                 reward: Callable[[_C_out,_A_out],Reward]) -> None:
+                 reward    : Callable[[_C_out,_A_out],Reward]) -> None:
         """Instantiate a LambdaSimulation.
 
         Args:
@@ -245,7 +272,7 @@ class LambdaSimulation(Simulation[_C_out, _A_out]):
             reward: A function that should return the reward for a context and action.
         """
 
-        contexts     : List[_C_out]           = []
+        contexts   : List[_C_out]           = []
         action_sets: List[Sequence[_A_out]] = []
         reward_sets: List[Sequence[Reward]] = []
 
@@ -279,16 +306,16 @@ class LambdaSimulation(Simulation[_C_out, _A_out]):
         return self._simulation.rewards(choices)
 
 class ShuffleSimulation(Simulation[_C_out, _A_out]):
-    """A simulation which created from an existing simulation by shuffling interactions.
+    """A simulation created from an existing simulation by shuffling interactions.
 
     Remarks:
         Shuffling is applied one time upon creation and after that interaction order is fixed.
         Shuffling does not change the original simulation's interaction order or copy the
-        original interactions in memory. Shuffling is guaranteed to be deterministic 
-        according any given to seed regardless of the local Python execution environment.
+        original interactions. Shuffling is guaranteed to be deterministic according to seed
+        regardless of the local Python execution environment.
     """
 
-    def __init__(self, simulation: Simulation[_C_out,_A_out], seed: Optional[int] = None):
+    def __init__(self, simulation: Simulation[_C_out,_A_out], seed: Optional[int]):
         """Instantiate a ShuffleSimulation
 
         Args:
@@ -296,10 +323,12 @@ class ShuffleSimulation(Simulation[_C_out, _A_out]):
             seed: The seed we wish to use in determining the shuffle order.
         """
 
-        cb_random.seed(seed)
+        if seed is None:
+            self._interactions = simulation.interactions
+        else:
+            self._interactions = cb_random.Random(seed).shuffle(simulation.interactions)
 
-        self._interactions = cb_random.shuffle(simulation.interactions)
-        self._rewards      = simulation.rewards
+        self._rewards = simulation.rewards
 
     @property
     def interactions(self) -> Sequence[Interaction[_C_out,_A_out]]:
@@ -320,7 +349,7 @@ class ShuffleSimulation(Simulation[_C_out, _A_out]):
 
         return self._rewards(choices)
 
-class ClassificationSimulation(Simulation[_C_out, _A_out]):
+class ClassificationSimulation(Simulation[_C_out, Tuple[int,...]]):
     """A simulation created from classifier data with features and labels.
 
     ClassificationSimulation turns labeled observations from a classification data set
@@ -338,7 +367,7 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
     """
 
     @staticmethod
-    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'ClassificationSimulation[Context,Action]':
+    def from_json(json_val:Union[str, Dict[str,Any]]) -> 'ClassificationSimulation[Context]':
         """Construct a ClassificationSimulation object from JSON.
 
         Args:
@@ -350,12 +379,13 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
 
         config = json.loads(json_val) if isinstance(json_val,str) else json_val
 
-        has_header  : bool                        = True
-        default_meta: Metadata[bool,bool,Encoder] = Metadata.default()
-        defined_meta: Dict[Any, Metadata]         = {}
+        has_header  : bool                = True
+        default_meta: FullMeta            = FullMeta()
+        defined_meta: Dict[Any, PartMeta] = {}
 
         if config["format"] == "openml":
-            return ClassificationSimulation.from_openml(config["id"])
+            with ExecutionContext.Logger.log(f"loading openml {config['id']}..."):
+                return ClassificationSimulation.from_openml(config["id"], config.get("md5_checksum", None))
 
         if config["format"] == "csv":
             location    : str           = config["location"]
@@ -368,11 +398,11 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
                 has_header = config["has_header"]
 
             if "column_default" in config:
-                default_meta = cast(Metadata[bool,bool,Encoder], Metadata.from_json(config["column_default"]))
+                default_meta =  FullMeta.from_json(config["column_default"])
 
             if "column_overrides" in config:
                 for key,value in config["column_overrides"].items():
-                    defined_meta[key] = Metadata.from_json(value)
+                    defined_meta[key] = PartMeta.from_json(value)
 
             return ClassificationSimulation.from_csv(
                 location     = location,
@@ -389,11 +419,11 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
                 has_header = config["has_header"]
 
             if "column_default" in config:
-                default_meta = cast(Metadata[bool,bool,Encoder], Metadata.from_json(config["column_default"]))
+                default_meta = FullMeta.from_json(config["column_default"])
 
             if "column_overrides" in config:
                 for key,value in config["column_overrides"].items():
-                    defined_meta[key] = Metadata.from_json(value)
+                    defined_meta[key] = PartMeta.from_json(value)
 
             return ClassificationSimulation.from_table(
                 table        = table,
@@ -405,53 +435,42 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
         raise Exception("We were unable to recognize the provided data format.")
 
     @staticmethod
-    def from_openml(data_id:int) -> 'ClassificationSimulation[Context,Action]':
+    def from_openml(data_id:int, md5_checksum:str = None) -> 'ClassificationSimulation[Context]':
         """Create a ClassificationSimulation from a given openml dataset id.
 
         Args:
             data_id: The unique identifier for a dataset stored on openml.
         """
 
-        with ExecutionContext.Logger.log(f"loading openml {data_id} meta... "):
+        openml_api_key = ExecutionContext.Config.openml_api_key
 
-            openml_api_key = ExecutionContext.CobaConfig.openml_api_key
+        data_description_url = f'https://www.openml.org/api/v1/json/data/{data_id}?api_key={openml_api_key}'
+        task_description_url = f'https://www.openml.org/api/v1/json/task/list/data_id/{data_id}?api_key={openml_api_key}'
+        feat_description_url = f'https://www.openml.org/api/v1/json/data/features/{data_id}?api_key={openml_api_key}'
 
-            with closing(urllib.request.urlopen(f'https://www.openml.org/api/v1/json/data/{data_id}?api_key={openml_api_key}')) as resp:
-                description = json.loads(resp.read())["data_set_description"]
+        descr = json.loads(''.join(HttpSource(data_description_url, '.json', None, 'descr').read()))["data_set_description"]
+        tasks = json.loads(''.join(HttpSource(task_description_url, '.json', None, 'tasks').read()))["tasks"]["task"]
+        types = json.loads(''.join(HttpSource(feat_description_url, '.json', None, 'types').read()))["data_features"]["feature"]
 
-            if description['status'] == 'deactivated':
-                raise Exception(f"Openml {data_id} has been deactivated. This is often due to flags on the data.")
+        if descr['status'] == 'deactivated':
+            raise Exception(f"Openml {data_id} has been deactivated. This is often due to flags on the data.")
 
-            with closing(urllib.request.urlopen(f'https://www.openml.org/api/v1/json/task/list/data_id/{data_id}?api_key={openml_api_key}')) as resp:
-                tasks = json.loads(resp.read())["tasks"]["task"]
+        if not any(task["task_type_id"] == 1 for task in tasks ):
+            raise Exception(f"Openml {data_id} does not appear to be a classification dataset")
 
-            if not any(task["task_type_id"] == 1 for task in tasks ):
-                raise Exception(f"Openml {data_id} does not appear to be a classification dataset")
+        defined_meta: Dict[str,PartMeta] = {}
 
-            with closing(urllib.request.urlopen(f'https://www.openml.org/api/v1/json/data/features/{data_id}?api_key={openml_api_key}')) as resp:
-                features = json.loads(resp.read())["data_features"]["feature"]
+        for tipe in types:
 
-            defined_meta: Dict[str,Metadata] = {}
+            defined_meta[tipe["name"]] = PartMeta(
+                ignore  = tipe["is_ignore"] == "true" or tipe["is_row_identifier"] == "true",
+                label   = tipe["is_target"] == "true",
+                encoder = NumericEncoder() if tipe['data_type'] == 'numeric' else FactorEncoder(tipe['nominal_value'], error_if_unknown=True)
+            )
 
-            for m in features:
+        csv_url = f"http://www.openml.org/data/v1/get_csv/{descr['file_id']}"
 
-                encoder: Encoder
-
-                if m['data_type'] == 'numeric':
-                    encoder = NumericEncoder()
-                else:
-                    encoder = FactorEncoder(m['nominal_value'],error_if_unknown=True)
-
-                defined_meta[m["name"]] = Metadata(
-                    ignore  = m["is_ignore"] == "true" or m["is_row_identifier"] == "true",
-                    label   = m["is_target"] == "true",
-                    encoder = encoder
-                )
-
-        file_id = description['file_id']
-        csv_url = f"http://www.openml.org/data/v1/get_csv/{file_id}"
-
-        return ClassificationSimulation.from_csv(csv_url, defined_meta=defined_meta)
+        return ClassificationSimulation.from_csv(csv_url, md5_checksum=md5_checksum, defined_meta=defined_meta)
 
     @staticmethod
     def from_csv(
@@ -460,8 +479,8 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
         md5_checksum: Optional[str] = None,
         csv_reader  : Callable[[Iterable[str]], Iterable[Sequence[str]]] = csv.reader, #type: ignore #pylance complains
         has_header  : bool = True,
-        default_meta: Metadata[bool,bool,Encoder] = Metadata.default(),
-        defined_meta: Dict[Any,Metadata] = {}) -> 'ClassificationSimulation[Context,Action]':
+        default_meta: FullMeta = FullMeta(),
+        defined_meta: Dict[Any,PartMeta] = {}) -> 'ClassificationSimulation[Context]':
         """Create a ClassificationSimulation given the location of a csv formatted dataset.
 
         Args:
@@ -474,50 +493,14 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
             column_metas: Keys are column name or index, values are meta objects that override the default values.
         """
 
-        cachename = f"{md5(location.encode('utf-8')).hexdigest()}.csv"
+        if not location.lower().startswith('http'):
+            source = DiskSource(location)
+        else: 
+            source = HttpSource(location, ".csv", md5_checksum, 'data')
 
-        is_cache =      cachename in ExecutionContext.FileCache 
-        is_disk  =  not location.lower().startswith('http') and not is_cache
-        is_http  =      location.lower().startswith('http') and not is_cache
+        csv_rows = list(csv_reader(source.read()))
 
-        if is_cache:
-            source         = "cache"
-            stream_manager = ExecutionContext.FileCache.get(cachename)
-        elif is_disk:
-            source         = "disk"
-            stream_manager = open(location, 'rb')
-        else:
-            source         = "http"
-            http_request   = urllib.request.Request(location, headers={'Accept-encoding':'gzip'})
-            stream_manager = urllib.request.urlopen(http_request)
-
-        with stream_manager as raw_stream:
-
-            with ExecutionContext.Logger.log(f'loading csv from {source}... ', end=''):
-                is_cache_gzip = False
-                is_disk_gzip  = is_disk and location.lower().endswith(".gz")
-                is_http_gzip  = is_http and cast(HTTPResponse, raw_stream).info().get('Content-Encoding') == "gzip"
-                is_gzip       = is_disk_gzip or is_http_gzip or is_cache_gzip
-
-                if is_gzip: cachename += ".gz"
-                if is_http: raw_stream = ExecutionContext.FileCache.put(cachename, raw_stream)
-
-                #When testing loading all bytes into memory at once was moderately faster on average.
-                #This does run the risk of causing problems though if the file is extremely large.
-                raw_bytes = gzip.decompress(raw_stream.read()) if is_gzip else raw_stream.read()
-
-                if md5_checksum is not None and md5_checksum != md5(raw_bytes).hexdigest():
-                    ExecutionContext.FileCache.rmv(cachename)
-                    raise Exception(
-                        "The dataset did not match the expected checksum. This could be the result of network "
-                        "errors or the file becoming corrupted. Please consider downloading the file again and if "
-                        "the error persists you may want to manually download and reference the file.")
-
-        csv_rows = csv_reader(raw_bytes.decode('utf-8').split("\n"))
-
-        del raw_bytes
-
-        with ExecutionContext.Logger.log('encoding csv in memory... ', end=''):
+        with ExecutionContext.Logger.log('encoding data... '):
             return ClassificationSimulation.from_table(csv_rows, label_col, has_header, default_meta, defined_meta)
 
     @staticmethod
@@ -525,8 +508,8 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
         table       : Iterable[Sequence[str]],
         label_col   : Union[None,str,int] = None,
         has_header  : bool = True,
-        default_meta: Metadata[bool,bool,Encoder] = Metadata.default(),
-        defined_meta: Dict[Any,Metadata] = {}) -> 'ClassificationSimulation[Context,Action]':
+        default_meta: FullMeta = FullMeta(),
+        defined_meta: Dict[Any, PartMeta] = {}) -> 'ClassificationSimulation[Context]':
         """Create a ClassifierSimulation from the rows contained in a csv formatted dataset.
 
         Args:
@@ -534,7 +517,7 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
             label_col: Either the column index or the header name for the label column.
             has_header: Indicates if the first row in the table contains column names
             default_meta: The default meta values for all columns unless explictly overridden with column_metas.
-            column_metas: Keys are column name or index, values are meta objects that override the default values.
+            defined_meta: Keys are column name or index, values are meta objects that override the default values.
         """
 
         # In theory we don't have to load the whole file up front. However, in practice,
@@ -564,12 +547,11 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
         def index(key: Union[int,str]):
             return header.index(key) if isinstance(key,str) else key
 
-        empty_meta = Metadata(None,None,None)
-        over_metas = defaultdict(lambda:empty_meta, { index(key):val for key,val in defined_meta.items() } )
+        over_metas = defaultdict(PartMeta, { index(key):val for key,val in defined_meta.items() } )
         metas      = [ default_meta.override(over_metas[i]) for i in range(n_col) ]
 
         if label_index is not None:
-            metas[label_index] = metas[label_index].override(Metadata(None,True,None))
+            metas[label_index] = metas[label_index].override(PartMeta(label=True))
 
         #after extensive testing I found that performing many loops with simple logic
         #was about 3 times faster than performing one or two loops with complex logic
@@ -592,11 +574,12 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
         for col, m in zip(columns, metas):
 
             encoding = label_encodings if m.label else feature_encodings
+            encoder  = m.encoder if m.encoder.is_fit else m.encoder.fit(col)
 
-            if isinstance(m.encoder, OneHotEncoder):
-                encoding.extend(list(zip(*m.encoder.fit_encode(col))))
+            if isinstance(encoder, OneHotEncoder):
+                encoding.extend(list(zip(*encoder.encode(col))))
             else:
-                encoding.append(m.encoder.fit_encode(col))
+                encoding.append(encoder.encode(col))
 
         #transform columns back into rows
         features = list(zip(*feature_encodings)) #type: ignore
@@ -608,7 +591,7 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
 
         return ClassificationSimulation(contexts, actions)
 
-    def __init__(self, features: Sequence[_C_out], labels: Sequence[_A_out]) -> None:
+    def __init__(self, features: Sequence[_C_out], labels: Sequence[Action]) -> None:
         """Instantiate a ClassificationSimulation.
 
         Args:
@@ -618,16 +601,17 @@ class ClassificationSimulation(Simulation[_C_out, _A_out]):
 
         assert len(features) == len(labels), "Mismatched lengths of features and labels"
 
-        action_set = tuple(set(labels))
+        action_set = list(set(labels))
 
-        contexts  = features
-        actions = list(repeat(action_set, len(contexts)))
-        rewards = OneHotEncoder(action_set,False,True).encode(labels)
+        contexts = features
+        actions  = list(repeat(OneHotEncoder(action_set).encode(action_set), len(contexts)))
+        rewards  = OneHotEncoder(action_set).encode(labels)
 
+        self._action_set = action_set
         self._simulation = MemorySimulation(contexts, actions, rewards)
 
     @property
-    def interactions(self) -> Sequence[Interaction[_C_out, _A_out]]:
+    def interactions(self) -> Sequence[Interaction[_C_out, Tuple[int,...]]]:
         """The interactions in this simulation.
 
         Remarks:
